import asyncio
import contextlib
import json
from collections import namedtuple
from enum import Enum
from typing import Awaitable, Dict, List, Optional, cast

import websockets
from discord.backoff import ExponentialBackoff

from . import log
from .rest_api import Track


__all__ = [
    "DiscordVoiceSocketResponses",
    "LavalinkEvents",
    "TrackEndReason",
    "LavalinkOutgoingOp",
    "PlayerState",
    "Stats",
    "Node",
    "get_node",
    "join_voice",
]

_nodes = {}  # type: Dict[Node, List[int]]


class DiscordVoiceSocketResponses(Enum):
    VOICE_STATE_UPDATE = "VOICE_STATE_UPDATE"
    VOICE_SERVER_UPDATE = "VOICE_SERVER_UPDATE"


class LavalinkIncomingOp(Enum):
    EVENT = "event"
    PLAYER_UPDATE = "playerUpdate"
    STATS = "stats"


class LavalinkOutgoingOp(Enum):
    VOICE_UPDATE = "voiceUpdate"
    DESTROY = "destroy"
    PLAY = "play"
    STOP = "stop"
    PAUSE = "pause"
    SEEK = "seek"
    VOLUME = "volume"


class LavalinkEvents(Enum):
    """
    An enumeration of the Lavalink Track Events.
    """

    TRACK_END = "TrackEndEvent"
    """The track playback has ended."""

    TRACK_EXCEPTION = "TrackExceptionEvent"
    """There was an exception during track playback."""

    TRACK_STUCK = "TrackStuckEvent"
    """Track playback got stuck during playback."""

    # Custom events
    TRACK_START = "TrackStartEvent"
    """This is a custom event generated by this library that is used to
    denote the start of a track.
    """

    QUEUE_END = "QueueEndEvent"
    """This is a custom event generated by this library to denote the
    end of all tracks in the queue.
    """


class TrackEndReason(Enum):
    """
    The reasons why track playback has ended.
    """

    FINISHED = "FINISHED"
    """The track reached the end, or the track itself ended with an
    exception.
    """

    LOAD_FAILED = "LOAD_FAILED"
    """The track failed to start, throwing an exception before
    providing any audio.
    """

    STOPPED = "STOPPED"
    """The track was stopped due to the player being stopped.
    """

    REPLACED = "REPLACED"
    """The track stopped playing because a new track started playing.
    """

    CLEANUP = "CLEANUP"
    """The track was stopped because the cleanup threshold for the
    audio player was reached.
    """


PlayerState = namedtuple("PlayerState", "position time")
MemoryInfo = namedtuple("MemoryInfo", "reservable used free allocated")
CPUInfo = namedtuple("CPUInfo", "cores systemLoad lavalinkLoad")


class Stats:
    def __init__(self, memory, players, active_players, cpu, uptime):
        self.memory = MemoryInfo(**memory)
        self.players = players
        self.active_players = active_players
        self.cpu_info = CPUInfo(**cpu)
        self.uptime = uptime


class Node:

    _is_shutdown = False  # type: bool

    def __init__(
        self, _loop, event_handler, voice_ws_func, host, password, port, rest, user_id, num_shards
    ):
        """
        Represents a Lavalink node.

        Parameters
        ----------
        _loop : asyncio.BaseEventLoop
            The event loop of the bot.
        event_handler
            Function to dispatch events to.
        voice_ws_func : typing.Callable
            Function that takes one argument, guild ID, and returns a websocket.
        host : str
            Lavalink player host.
        password : str
            Password for the Lavalink player.
        port : int
            Port of the Lavalink player event websocket.
        rest : int
            Port for the Lavalink REST API.
        user_id : int
            User ID of the bot.
        num_shards : int
            Number of shards to which the bot is currently connected.
        ready : asyncio.Event
            Set when the connection is up and running, unset when not.
        """
        self.loop = _loop
        self.event_handler = event_handler
        self.voice_ws_func = voice_ws_func
        self.host = host
        self.port = port
        self.rest = rest
        self.password = password
        self.headers = self._get_connect_headers(self.password, user_id, num_shards)

        self.ready = asyncio.Event()

        self._ws = None
        self._listener_task = None

        self._queue = []
        self._players = set()

    async def connect(self, timeout=None):
        """
        Connects to the Lavalink player event websocket.

        Parameters
        ----------
        timeout : int
            Time after which to timeout on attempting to connect to the Lavalink websocket,
            ``None`` is considered never.

        Raises
        ------
        asyncio.TimeoutError
            If the websocket failed to connect after the given time.
        """
        self._is_shutdown = False

        combo_uri = "ws://{}:{}".format(self.host, self.rest)
        uri = "ws://{}:{}".format(self.host, self.port)

        log.debug(
            "Lavalink WS connecting to %s or %s with headers %s", combo_uri, uri, self.headers
        )

        tasks = tuple({self._multi_try_connect(u) for u in (combo_uri, uri)})

        for task in asyncio.as_completed(tasks, timeout=timeout):
            with contextlib.suppress(Exception):
                if await cast(Awaitable[Optional[websockets.WebSocketClientProtocol]], task):
                    break

        _nodes[self] = []

        log.debug("Creating Lavalink WS listener.")
        self._listener_task = self.loop.create_task(self.listener())

        for data in self._queue:
            await self.send(data)

        self.ready.set()

    async def wait_until_ready(self, timeout: Optional[float] = None):
        await asyncio.wait_for(self.ready.wait(), timeout=timeout)

    @staticmethod
    def _get_connect_headers(password, user_id, num_shards):
        return {"Authorization": password, "User-Id": user_id, "Num-Shards": num_shards}

    @property
    def lavalink_major_version(self):
        assert self._ws, "not connected"
        return self._ws.response_headers.get("Lavalink-Major-Version")

    async def _multi_try_connect(self, uri):
        backoff = ExponentialBackoff()
        attempt = 1

        while self._is_shutdown is False and (self._ws is None or not self._ws.open):
            try:
                ws = self._ws = await websockets.connect(uri, extra_headers=self.headers)
                return ws
            except OSError:
                delay = backoff.delay()
                log.debug("Failed connect attempt %s, retrying in %s", attempt, delay)
                await asyncio.sleep(delay)
                attempt += 1
            except websockets.InvalidStatusCode:
                return None

    async def listener(self):
        """
        Listener task for receiving ops from Lavalink.
        """
        while self._ws.open and self._is_shutdown is False:
            try:
                data = json.loads(await self._ws.recv())
            except websockets.ConnectionClosed:
                break

            raw_op = data.get("op")
            try:
                op = LavalinkIncomingOp(raw_op)
            except ValueError:
                log.debug("Received unknown op: %s", data)
            else:
                log.debug("Received known op: %s", data)
                self.loop.create_task(self._handle_op(op, data))

        self.ready.clear()
        log.debug("Listener exited: ws %s SHUTDOWN %s.", self._ws.open, self._is_shutdown)
        self.loop.create_task(self._reconnect())

    async def _handle_op(self, op: LavalinkIncomingOp, data):
        if op == LavalinkIncomingOp.EVENT:
            try:
                event = LavalinkEvents(data.get("type"))
            except ValueError:
                log.debug("Unknown event type: %s", data)
            else:
                self.event_handler(op, event, data)
        elif op == LavalinkIncomingOp.PLAYER_UPDATE:
            state = PlayerState(**data.get("state"))
            self.event_handler(op, state, data)
        elif op == LavalinkIncomingOp.STATS:
            stats = Stats(
                memory=data.get("memory"),
                players=data.get("players"),
                active_players=data.get("playingPlayers"),
                cpu=data.get("cpu"),
                uptime=data.get("uptime"),
            )
            self.event_handler(op, stats, data)

    async def _reconnect(self):
        self.ready.clear()

        if self._is_shutdown is True:
            log.debug("Shutting down Lavalink WS.")
            return

        log.debug("Attempting Lavalink WS reconnect.")
        try:
            await self.connect()
        except asyncio.TimeoutError:
            log.debug("Failed to reconnect, please reinitialize lavalink when ready.")
        else:
            log.debug("Reconnect successful.")

    async def disconnect(self):
        """
        Shuts down and disconnects the websocket.
        """
        self._is_shutdown = True
        self.ready.clear()

        if self._ws is not None and self._ws.open:
            await self._ws.close()

        while self._players:
            await self._players.pop().destroy()

        if _nodes.pop(self, None):
            log.debug("Shutdown Lavalink WS.")

    async def send(self, data):
        if self._ws is None or not self._ws.open:
            self._queue.append(data)
        else:
            log.debug("Sending data to Lavalink: %s", data)
            await self._ws.send(json.dumps(data))

    async def send_lavalink_voice_update(self, guild_id, session_id, event):
        await self.send(
            {
                "op": LavalinkOutgoingOp.VOICE_UPDATE.value,
                "guildId": str(guild_id),
                "sessionId": session_id,
                "event": event,
            }
        )

    async def destroy_guild(self, guild_id: int):
        await self.send({"op": LavalinkOutgoingOp.DESTROY.value, "guildId": str(guild_id)})

    # Player commands
    async def stop(self, guild_id: int):
        await self.send({"op": LavalinkOutgoingOp.STOP.value, "guildId": str(guild_id)})
        self.event_handler(
            LavalinkIncomingOp.EVENT, LavalinkEvents.QUEUE_END, {"guildId": str(guild_id)}
        )

    async def play(self, guild_id: int, track: Track):
        await self.send(
            {
                "op": LavalinkOutgoingOp.PLAY.value,
                "guildId": str(guild_id),
                "track": track.track_identifier,
            }
        )
        self.event_handler(
            LavalinkIncomingOp.EVENT,
            LavalinkEvents.TRACK_START,
            {"guildId": str(guild_id), "track": track},
        )

    async def pause(self, guild_id, paused):
        await self.send(
            {"op": LavalinkOutgoingOp.PAUSE.value, "guildId": str(guild_id), "pause": paused}
        )

    async def volume(self, guild_id: int, _volume: int):
        await self.send(
            {"op": LavalinkOutgoingOp.VOLUME.value, "guildId": str(guild_id), "volume": _volume}
        )

    async def seek(self, guild_id: int, position: int):
        await self.send(
            {"op": LavalinkOutgoingOp.SEEK.value, "guildId": str(guild_id), "position": position}
        )


def get_node(guild_id: int) -> Node:
    """
    Gets a node based on a guild ID, useful for noding separation. If the
    guild ID does not already have a node association, the least used
    node is returned. Skips over nodes that are not yet ready.

    Parameters
    ----------
    guild_id : int

    Raises
    ------
    IndexError
        If no Nodes have been instantiated yet.

    Returns
    -------
    Node
    """
    guild_count = 1e10
    least_used = None

    if not _nodes:
        raise IndexError("no Nodes have been instantiated")

    for node, guild_ids in _nodes.items():
        if not node.ready.is_set():
            continue
        elif len(guild_ids) < guild_count:
            guild_count = len(guild_ids)
            least_used = node

        if guild_id in guild_ids:
            return node

    if least_used is None:
        raise IndexError("No nodes found.")

    _nodes[least_used].append(guild_id)
    return least_used


async def join_voice(guild_id: int, channel_id: int):
    """
    Joins a voice channel by ID's.

    Parameters
    ----------
    guild_id : int
    channel_id : int
    """
    node = get_node(guild_id)
    voice_ws = node.voice_ws_func(guild_id)
    await voice_ws.voice_state(guild_id, channel_id)


async def disconnect():
    nodes = list(_nodes.keys())
    for node in nodes:
        await node.disconnect()



import logging

log = logging.getLogger("red.core.lavalink")

try:
    from .lavalink import *
    from .node import Node, LavalinkEvents, TrackEndReason, PlayerState, Stats
    from .player_manager import *
    from .rest_api import Track
    from . import utils
except ImportError:
    pass

__version__ = "0.2.2"



import asyncio
from random import randrange
from typing import Optional

import discord

from . import log
from . import node
from .rest_api import Track, RESTClient

__all__ = ["players", "user_id", "channel_finder_func", "connect", "get_player", "Player"]

players = []
user_id = None
channel_finder_func = lambda channel_id: None

_voice_states = {}
_players_dict = {}


class Player(RESTClient):
    """
    The Player class represents the current state of playback.
    It also is used to control playback and queue tracks.

    The existence of this object guarantees that the bot is connected
    to a voice voice channel.

    Attributes
    ----------
    channel: discord.VoiceChannel
        The channel the bot is connected to.
    queue : list of Track
    position : int
        The seeked position in the track of the current playback.
    current : Track
    repeat : bool
    shuffle : bool
    """

    def __init__(self, node_: node.Node, channel: discord.VoiceChannel):
        super().__init__(node_)
        self.channel = channel

        self.queue = []
        self.position = 0
        self.current = None  # type: Track
        self._paused = False
        self.repeat = False
        self.shuffle = False

        self._volume = 100

        self._is_playing = False
        self._metadata = {}
        self._node = node_

        if channel.guild.id in _players_dict:
            raise ValueError("player for guild {} already exists".format(channel.guild.id))

        _players_dict[channel.guild.id] = self
        players.append(self)
        self._node._players.add(self)

    @property
    def is_playing(self) -> bool:
        """
        Current status of playback
        """
        return self._is_playing and not self._paused

    @property
    def paused(self) -> bool:
        """
        Player's paused state.
        """
        return self._paused

    @property
    def volume(self) -> int:
        """
        The current volume.
        """
        return self._volume

    @property
    def ready(self) -> bool:
        """
        Whether the underlying node is ready for requests.
        """
        return self._node.ready.is_set()

    async def wait_until_ready(
        self, timeout: Optional[float] = None, no_raise: bool = False
    ) -> bool:
        """
        Waits for the underlying node to become ready.

        If no_raise is set, returns false when a timeout occurs instead of propogating TimeoutError.
        A timeout of None means to wait indefinitely.
        """
        if self._node.ready.is_set():
            return True

        try:
            return await self._node.wait_until_ready(timeout=timeout)
        except asyncio.TimeoutError:
            if no_raise:
                return False
            else:
                raise

    async def connect(self):
        """
        Connects to the voice channel associated with this Player.
        """
        await node.join_voice(self.channel.guild.id, self.channel.id)

    async def move_to(self, channel: discord.VoiceChannel):
        """
        Moves this player to a voice channel.

        Parameters
        ----------
        channel : discord.VoiceChannel
        """
        if channel.guild != self.channel.guild:
            raise TypeError("Cannot move to a different guild.")

        self.channel = channel
        await self.connect()

    async def disconnect(self):
        """
        Disconnects this player from it's voice channel.
        """
        guild_id = self.channel.guild.id
        voice_ws = self._node.voice_ws_func(guild_id)

        if not voice_ws.closed:
            await voice_ws.voice_state(guild_id, None)

        await self._node.destroy_guild(guild_id)
        await self.close()

    async def destroy(self):
        if self in players:
            players.remove(self)

        if _players_dict.get(self.channel.guild.id) is self:
            _players_dict.pop(self.channel.guild.id)

        self._node._players.discard(self)

        await self.disconnect()

    def store(self, key, value):
        """
        Stores a metadata value by key.
        """
        self._metadata[key] = value

    def fetch(self, key, default=None):
        """
        Returns a stored metadata value.

        Parameters
        ----------
        key
            Key used to store metadata.
        default
            Optional, used if the key doesn't exist.
        """
        return self._metadata.get(key, default)

    async def _handle_event(self, event: node.LavalinkEvents, extra):
        """
        Handles various Lavalink Events.

        If the event is TRACK END, extra will be TrackEndReason.

        If the event is TRACK EXCEPTION, extra will be the string reason.

        If the event is TRACK STUCK, extra will be the threshold ms.

        Parameters
        ----------
        event : node.LavalinkEvents
        extra
        """
        if event == node.LavalinkEvents.TRACK_END:
            if extra == node.TrackEndReason.FINISHED:
                await self.play()
            else:
                self._is_playing = False

    async def _handle_player_update(self, state: node.PlayerState):
        """
        Handles player updates from lavalink.

        Parameters
        ----------
        state : websocket.PlayerState
        """
        if state.position > self.position:
            self._is_playing = True
        self.position = state.position

    # Play commands
    def add(self, requester: discord.User, track: Track):
        """
        Adds a track to the queue.

        Parameters
        ----------
        requester : discord.User
            User who requested the track.
        track : Track
            Result from any of the lavalink track search methods.
        """
        track.requester = requester
        self.queue.append(track)

    async def play(self):
        """
        Starts playback from lavalink.
        """
        if self.repeat and self.current is not None:
            self.queue.append(self.current)

        self.current = None
        self.position = 0
        self._paused = False

        if not self.queue:
            await self.stop()
        else:
            self._is_playing = True
            if self.shuffle:
                track = self.queue.pop(randrange(len(self.queue)))
            else:
                track = self.queue.pop(0)

            self.current = track
            log.debug("Assigned current.")
            await self._node.play(self.channel.guild.id, track)

    async def stop(self):
        """
        Stops playback from lavalink.

        .. important::

            This method will clear the queue.
        """
        await self._node.stop(self.channel.guild.id)
        self.queue = []
        self.current = None
        self.position = 0
        self._paused = False

    async def skip(self):
        """
        Skips to the next song.
        """
        await self.play()

    async def pause(self, pause: bool = True):
        """
        Pauses the current song.

        Parameters
        ----------
        pause : bool
            Set to ``False`` to resume.
        """
        self._paused = pause
        await self._node.pause(self.channel.guild.id, pause)

    async def set_volume(self, volume: int):
        """
        Sets the volume of Lavalink.

        Parameters
        ----------
        volume : int
            Between 0 and 150
        """
        self._volume = max(min(volume, 150), 0)
        await self._node.volume(self.channel.guild.id, self.volume)

    async def seek(self, position: int):
        """
        If the track allows it, seeks to a position.

        Parameters
        ----------
        position : int
            Between 0 and track length.
        """
        if self.current.seekable:
            position = max(min(position, self.current.length), 0)
            await self._node.seek(self.channel.guild.id, position)


async def connect(channel: discord.VoiceChannel) -> Player:
    """
    Connects to a discord voice channel.

    Parameters
    ----------
    channel

    Returns
    -------
    Player
        The created Player object.
    """
    if _already_in_guild(channel):
        p = get_player(channel.guild.id)
        await p.move_to(channel)
    else:
        ws = node.get_node(channel.guild.id)
        p = Player(ws, channel)
        await p.connect()
    return p


def _already_in_guild(channel: discord.VoiceChannel) -> bool:
    for p in players:
        if p.channel.guild == channel.guild:
            return True
    return False


def get_player(guild_id: int) -> Player:
    """
    Gets a Player object from a guild ID.

    Parameters
    ----------
    guild_id : int
        Discord guild ID.

    Returns
    -------
    Player

    Raises
    ------
    KeyError
        If that guild does not have a Player, e.g. is not connected to any
        voice channel.
    """
    if guild_id in _players_dict:
        return _players_dict[guild_id]
    raise KeyError("No such player for that guild.")


def _ensure_player(channel_id: int):
    channel = channel_finder_func(channel_id)
    if channel is not None:
        try:
            p = get_player(channel.guild.id)
        except KeyError:
            log.debug("Received voice channel connection without a player.")
            node_ = node.get_node(channel.guild.id)
            p = Player(node_, channel)
        return p, channel


async def _remove_player(guild_id: int):
    try:
        p = get_player(guild_id)
    except KeyError:
        pass
    else:
        await p.destroy()


async def on_socket_response(data):
    raw_event = data.get("t")
    try:
        event = node.DiscordVoiceSocketResponses(raw_event)
    except ValueError:
        return

    log.debug("Received Discord WS voice response: %s", data)

    guild_id = data["d"]["guild_id"]
    if guild_id not in _voice_states:
        _voice_states[guild_id] = {}

    if event == node.DiscordVoiceSocketResponses.VOICE_SERVER_UPDATE:
        # Connected for the first time
        socket_event_data = data["d"]

        _voice_states[guild_id].update({"guild_id": guild_id, "event": socket_event_data})
    elif event == node.DiscordVoiceSocketResponses.VOICE_STATE_UPDATE:
        channel_id = data["d"]["channel_id"]
        event_user_id = int(data["d"].get("user_id"))

        if event_user_id != user_id:
            return

        if channel_id is None:
            # We disconnected
            log.debug("Received voice disconnect from discord, removing player.")
            _voice_states[guild_id] = {}
            await _remove_player(int(guild_id))

        else:
            # After initial connection, get session ID
            p, channel = _ensure_player(int(channel_id))
            if channel != p.channel:
                p.channel = channel

        session_id = data["d"]["session_id"]
        _voice_states[guild_id]["session_id"] = session_id
    else:
        return

    if len(_voice_states[guild_id]) == 3:
        node_ = node.get_node(int(guild_id))
        await node_.send_lavalink_voice_update(**_voice_states[guild_id])


async def disconnect():
    """
    Disconnects all players.
    """
    for p in players:
        await p.destroy()
    log.debug("Disconnected players.")



from typing import Tuple

from aiohttp import ClientSession
from collections import namedtuple
from enum import Enum

from . import log

from urllib.parse import quote

__all__ = ["Track", "RESTClient", "LoadType", "PlaylistInfo"]


PlaylistInfo = namedtuple("PlaylistInfo", "name selectedTrack")


class LoadType(Enum):
    """
    The result type of a loadtracks request

    Attributes
    ----------
    TRACK_LOADED
    TRACK_LOADED
    PLAYLIST_LOADED
    SEARCH_RESULT
    NO_MATCHES
    LOAD_FAILED
    """

    TRACK_LOADED = "TRACK_LOADED"
    PLAYLIST_LOADED = "PLAYLIST_LOADED"
    SEARCH_RESULT = "SEARCH_RESULT"
    NO_MATCHES = "NO_MATCHES"
    LOAD_FAILED = "LOAD_FAILED"


class Track:
    """
    Information about a Lavalink track.

    Attributes
    ----------
    requester : discord.User
        The user who requested the track.
    track_identifier : str
        Track identifier used by the Lavalink player to play tracks.
    seekable : bool
        Boolean determining if seeking can be done on this track.
    author : str
        The author of this track.
    length : int
        The length of this track in milliseconds.
    is_stream : bool
        Determines whether Lavalink will stream this track.
    position : int
        Current seeked position to begin playback.
    title : str
        Title of this track.
    uri : str
        The playback url of this track.
    """

    def __init__(self, data):
        self.requester = None

        self.track_identifier = data.get("track")
        self._info = data.get("info", {})
        self.seekable = self._info.get("isSeekable", False)
        self.author = self._info.get("author")
        self.length = self._info.get("length", 0)
        self.is_stream = self._info.get("isStream", False)
        self.position = self._info.get("position")
        self.title = self._info.get("title")
        self.uri = self._info.get("uri")

    @property
    def thumbnail(self):
        """Optional[str]: Returns a thumbnail URL for YouTube tracks."""
        if "youtube" in self.uri and "identifier" in self._info:
            return "https://img.youtube.com/vi/{}/mqdefault.jpg".format(self._info["identifier"])


class LoadResult:
    """
    The result of a load_tracks request.

    Attributes
    ----------
    load_type : LoadType
        The result of the loadtracks request
    playlist_info : Optional[PlaylistInfo]
        The playlist information detected by Lavalink
    tracks : Tuple[Track, ...]
        The tracks that were loaded, if any
    """

    def __init__(self, data):
        self._raw = data
        self.load_type = LoadType(data["loadType"])

        if data.get("playlistInfo"):
            self.playlist_info = PlaylistInfo(**data["playlistInfo"])
        else:
            self.playlist_info = None

        self.tracks = tuple(Track(t) for t in data["tracks"])


class RESTClient:
    """
    Client class used to access the REST endpoints on a Lavalink node.
    """

    def __init__(self, node):
        self._node = node
        self._session = ClientSession(loop=node.loop)
        self._uri = "http://{}:{}/loadtracks?identifier=".format(node.host, node.rest)
        self._headers = {"Authorization": node.password}

    async def load_tracks(self, query) -> LoadResult:
        """
        Executes a loadtracks request. Only works on Lavalink V3.

        Parameters
        ----------
        query : str

        Returns
        -------
        LoadResult
        """
        url = self._uri + quote(str(query))

        async with self._session.get(url, headers=self._headers) as resp:
            data = await resp.json(content_type=None)

        assert type(data) is dict, "Lavalink V3 is required for this method"
        return LoadResult(data)

    async def get_tracks(self, query) -> Tuple[Track, ...]:
        """
        Gets tracks from lavalink.

        Parameters
        ----------
        query : str

        Returns
        -------
        Tuple[Track, ...]
        """
        url = self._uri + quote(str(query))

        async with self._session.get(url, headers=self._headers) as resp:
            data = await resp.json(content_type=None)

        tracks = data["tracks"] if type(data) is dict else data
        return tuple(Track(t) for t in tracks)

    async def search_yt(self, query) -> Tuple[Track, ...]:
        """
        Gets track results from YouTube from Lavalink.

        Parameters
        ----------
        query : str

        Returns
        -------
        list of Track
        """
        return await self.get_tracks("ytsearch:{}".format(query))

    async def search_sc(self, query) -> Tuple[Track, ...]:
        """
        Gets track results from SoundCloud from Lavalink.

        Parameters
        ----------
        query : str

        Returns
        -------
        list of Track
        """
        return await self.get_tracks("scsearch:{}".format(query))

    async def close(self):
        await self._session.close()
        log.debug("Closed REST session.")
